/** Pikku esimerkki yhdenaikaisuuden aiheuttamasta kilpa-ajosta eli
 *  race conditionista. Summaus globaaliin muuttujaan vaatii useita
 *  konekielisiä käskyjä, joiden välissä voi tulla kellokeskeytys!
 *
 *  Ongelma on tässä hallittu keskinäisesti poissulkevalla (mutual
 *  exclusion, mutex) lukolla: Lukitus vaatii viimekädessä
 *  käyttöjärjestelmän palveluita, koska sillä halutaan vaikuttaa
 *  vuoronnusjärjestykseen ja säikeiden (tai prosessien)
 *  tilasiirtymiin. Jos lukko on jollain muulla varattuna, uusi
 *  pyytäjä on otettava pois normaalista prosessikierrosta ja
 *  "blokattava" eli laitettava odottamaan lukon vapauttamista.
 *
 *  HUOM: Esimerkkien lyhentämiseksi näissä ei tarkisteta,
 *  onnistuivatko operaatiot. Oikeassa ohjelmassa AINA tarkistettava
 *  esim. onnistuiko säikeen luonti vai ei, ja käsiteltävä
 *  epäonnistuminen!!
 */

#include <stdio.h>
#include <stdint.h>
#include <inttypes.h>
#include <pthread.h>

#define N 100000000
uint64_t summa = 0;

pthread_mutex_t mymutex = PTHREAD_MUTEX_INITIALIZER;

/** Nyt on synkronoitu lukolla, mutta suorituskyvyn isku on aika
 *  suuri. Johtopäätös: Älä laske summaa näin, jos mahdollista (mutta
 *  muista lukottaa mikä tahansa jaettua resurssia käsittelevä koodi,
 *  josta voi aiheutua kilpa-ajotilanne!!)
 */
void * saikeen_koodi(void *v) {
    int i;
    for (i = 1; i <= N; i++) {
        pthread_mutex_lock(&mymutex);
        summa++;
        pthread_mutex_unlock(&mymutex);
    }
    return NULL;
}

int main(int argc, char *argv[]) {
    pthread_t saieA, saieB;

    pthread_create(&saieA, NULL, saikeen_koodi, NULL);
    pthread_create(&saieB, NULL, saikeen_koodi, NULL);
     
    pthread_join(saieA, NULL);
    pthread_join(saieB, NULL);

    printf("Summa on %" PRId64 " - pittäis olla %" PRId64 "\n", summa, (uint64_t) 2*N);
     
    return 0;
}
